#ifndef _OBJCONVERTER2_H
#define _OBJCONVERTER2_H

#include "SimpleVec.h"
#include <string>
#include <vector>
#include <map>


struct Material
{
	float rgb[3];
	int texture;
};

class Geometry;

static std::string model_path("data/city/model/procedural_pompeii_area6_hires/");


typedef std::vector<SimpleVec> VertexArray;
typedef std::pair<float, float> Texcoord;
typedef std::map<std::string, Material *> MaterialTable;
typedef std::map<std::string, int> TextureTable;
typedef std::vector<Texcoord> TexcoordArray;
typedef std::vector<std::string> TextureArray;

class ogzstream;

/** Converts obj format into objects readable by our format
*/
class ObjConverter2
{
public:

	ObjConverter2();

	bool Convert(const std::string &inputFilename, 
		         const std::string &outputFilename);
	
	bool Convert2(const std::vector<std::string> &filenames, 
		          const std::string &outputFilename);

	bool LoadMaterials(const std::string &mtlFilename);

	~ObjConverter2();

protected:

	struct Geometry
	{
		Geometry(): 
	    mVertices(NULL),
		mNormals(NULL),
		mTexcoords(NULL), 
		mMaterial(NULL), 
		mVertexCount(0), 
		mTexcoordCount(0)
		{}

		~Geometry()
		{
			delete [] mVertices;
			delete [] mNormals;
			delete [] mTexcoords;
		}

		SimpleVec *mVertices;
		SimpleVec *mNormals;
		Texcoord *mTexcoords;
		Material *mMaterial;

		int mVertexCount;
		int mTexcoordCount;
	};

typedef std::vector<Geometry *> GeometryArray;

	void LoadShape(const VertexArray &faceVertices,
		           const VertexArray &faceNormals,
				   const std::vector<Texcoord> &faceTexcoords, 
				   Material *mat);

	void WriteGeometry(ogzstream &str, Geometry *geom);

	bool ReadFile(const std::string &inputFilename);
	bool WriteFile(const std::string &outputFilename);


	MaterialTable mMaterialTable;
	TextureTable mTextureTable;
	
	TextureArray mTextures;
	GeometryArray mGeometry;

	int mNumShapes;
};



#endif
